"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const AUTOMATA_REDUCER_DEFAULTS = {
    key: 'state',
    toStandardAction: identity,
    operator: identity
};
function createAutomataReducer(automata, init, opts) {
    if (isStringOrNumber(opts)) {
        return createAutomataReducer(automata, init, { key: opts });
    }
    const { key, toStandardAction, operator } = Object.assign({}, AUTOMATA_REDUCER_DEFAULTS, opts);
    const withDefaultAutomataState = withDefaultEntry(key, init);
    const withDefaultState = operator(withDefaultAutomataState);
    const _automata = preprocess(automata, key, withDefaultAutomataState, operator);
    return function (previous, event) {
        const { type } = toStandardAction(event) || {};
        const reducer = _automata[type] || withDefaultState;
        return reducer(previous, event);
    };
}
exports.default = createAutomataReducer;
function preprocess(spec, key, withDefaultAutomataState, operator) {
    const _automata = Object.keys(spec).reduce(function (automata, state) {
        const events = spec[state];
        return Object.keys(events).reduce(function (automata, type) {
            const reducers = []
                .concat(events[type])
                .map(withAutomataReducer);
            automata[type] = Object.assign({}, automata[type], { [state]: reducers });
            return automata;
        }, automata);
    }, {});
    return Object.keys(_automata).reduce(function (eventTypeToReducerMap, type) {
        const stateToReducersMap = _automata[type];
        eventTypeToReducerMap[type] = operator(reducer);
        return eventTypeToReducerMap;
        function reducer(previous = {}, event = {}) {
            let state = withDefaultAutomataState(previous);
            const reducers = stateToReducersMap[state[key]];
            if (reducers) {
                let i = reducers.length;
                while (i--) {
                    state = reducers[i](state, event);
                }
            }
            return state;
        }
    }, {});
    function withAutomataReducer(reducer) {
        return !isStringOrNumber(reducer)
            ? reducer
            : function (state) {
                return state && state[key] === reducer
                    ? state
                    : Object.assign({}, state, { [key]: reducer });
            };
    }
}
function withDefaultEntry(key, value) {
    return function (map = {}) {
        return (map[key] ? map : Object.assign({}, map, { [key]: value }));
    };
}
function isStringOrNumber(val) {
    const t = typeof (val && val.valueOf());
    return t === 'string' || t === 'number';
}
function identity(v) {
    return v;
}
